from typing import Optional

from pydantic import BaseModel


class BaseTranslationBuilder(BaseModel):
    key: str
    linked_to_key: Optional[str] = None
    name: str

    @classmethod
    def link_to(cls, key: str) -> "BaseTranslationBuilder":
        return cls(
            linked_to_key=key
        )



from typing import Callable, Any, Awaitable

from aiogram import BaseMiddleware, Bot
from aiogram.types import Message, Update, User


class TranslationMiddleware(BaseMiddleware):
    def __init__(self, translator: "Translator", extract_language_function: Callable[[User, Bot], Awaitable[str]], key: str = "language") -> None:
        self._translator = translator
        self._key = key
        self.extract_language_function = extract_language_function

    async def __call__(
        self,
        handler: Callable[[Update, dict[str, Any]], Awaitable[Any]],
        event: Update,
        data: dict[str, Any]
    ) -> Any:
        update = event.callback_query or event.message or event.channel_post\
                or event.chat_member or event.edited_message or event.chat_join_request\
                or event.chosen_inline_result or event.edited_channel_post\
                or event.my_chat_member or event.poll_answer.user or event.shipping_query\
                or event.pre_checkout_query
        user = update.from_user
        if user:
            code = await self.extract_language_function(user, data["bot"])
        else:
            code = self._translator.get_default_key()
        data[self._key] = self._translator.get_translation(code)
        return await handler(event, data)


from typing import Union, Callable, Awaitable

from aiogram import Dispatcher, Bot
from aiogram.types import User

from aiogram_translation.middlewares import TranslationMiddleware
from aiogram_translation.models import BaseTranslationBuilder
from aiogram_translation.errors import InvalidTranslation, InvalidDefaultTranslation


async def extract_language_from_user(user: User, bot: Bot) -> str:
    return user.language_code


class Translator:
    def __init__(self,
                 default_language_key: str = None,
                 translations: list[BaseTranslationBuilder] = None,
                 extract_language_function: Callable[[User, Bot], Awaitable[str]] = extract_language_from_user):
        self._default_language_key = default_language_key
        self._translations = {}
        self.extract_language_function = extract_language_function
        if translations:
            self.include(translations)

    def include(self, translations: Union[BaseTranslationBuilder, list[BaseTranslationBuilder]]):
        if isinstance(translations, list):
            for translation in translations:
                self._translations[translation.linked_to_key or translation.key] = translation
        else:
            self._translations[translations.linked_to_key or translations.key] = translations

    def exclude(self, key: str):
        self._translations.pop(key)

    def set_default(self, key: str | None):
        self._default_language_key = key

    def get_translation(self, language_key: str) -> BaseTranslationBuilder:
        _translation = self._translations.get(language_key, None)
        if not _translation:
            if self._default_language_key:
                _translation = self._translations.get(self.get_default_key(), None)
                _translation.linked_to_key = language_key
            else:
                raise InvalidTranslation(
                    f'There is no translation with key \"{language_key}\" and default translation is not set.')
        return _translation

    def get_default_key(self):
        _translation = self._translations.get(self._default_language_key, None)
        if not _translation:
            raise InvalidDefaultTranslation(
                f'There is no translation with key \"{self._default_language_key}\" that is set by default.')
        return self._default_language_key

    def register(self, dp: Dispatcher, key: str = "language"):
        middleware = TranslationMiddleware(self, key=key, extract_language_function=self.extract_language_function)

        dp.update.middleware(middleware)
